//---------------------------------------------------------------------------
// Flight Simulator X - Shader Effect Files
// Copyright (c) 2006, Microsoft Corporation
//---------------------------------------------------------------------------

#include <Common.fxh>
#include <MaterialDecl.fxh>
#include <D3D9Texture.fxh>
#include <FuncLibrary.fxh>

#define vLightDir vSunVectorWorld

// Potential textures
texture Water_BaseTexture               : MATERIAL_BASE_TEXTURE;
texture Water_EnvTexture                : MATERIAL_ENVIRONMENT_TEXTURE;
texture Water_BumpTexture               : MATERIAL_BUMP_TEXTURE;

// Global variables
float3 vCamEyePoint : VIEWPOSITION;

float fTimeScale        : WATER_BUMP_TIME_SCALE;             
float fBumpScale        : WATER_BUMP_SCALE;          
float fBumpUVScale      : WATER_BUMP_UV_SCALE;       
float fBumpU            : WATER_BUMP_DISTANCE_SCALE; 
float fBumpV            : WATER_BUMP_LIGHT_SCALE;    
float fFresnelFactorMin : WATER_FRESNEL_FACTOR_MIN;  
float fFresnelFactorMax : WATER_FRESNEL_FACTOR_MAX;  
float fSpecularPower    : SPECULAR_POWER;            
float fSpecularBoost    : SPECULAR_BOOST;            
float fSpecularBlend    : SPECULAR_BLEND;            
float fBumpRotation     : WATER_BUMP_ROTATION;       
float fBumpTimeX1       : WATER_BUMP_TIME_X1;        
float fBumpTimeY1       : WATER_BUMP_TIME_Y1;        
float fBumpTimeX2       : WATER_BUMP_TIME_X2;        
float fBumpTimeY2       : WATER_BUMP_TIME_Y2;        
float fBumpTimeScale2   : WATER_BUMP_TIME_SCALE_2;   

// Fog
float   g_fFogDensity            : FOG_DENSITY;
float4  g_cFog                   : FOG_COLOR;

// Wind Information
float   g_fWindSpeed             : WIND_SPEED;
float   g_fWindDirection         : WIND_DIRECTION;
float   g_fViewElevation         : VIEW_ELEVATION;

// Samples used by this shader
sampler Water_BaseSampler = sampler_state
{
    Texture       = (Water_BaseTexture);
    AddressU      = Clamp;      
    AddressV      = Clamp;
    MinFilter     = (State_MinFilter);
    MagFilter     = (State_MagFilter);
    MipFilter     = (State_MipFilter);
    MipMapLodBias = (State_MipMapLodBias);
};

sampler Water_BumpSampler = sampler_state
{
    Texture       = (Water_BumpTexture);
    AddressU      = Wrap;
    AddressV      = Wrap;
    AddressW      = Wrap;
    MinFilter     = D3DTEXF_ANISOTROPIC;
    MagFilter     = D3DTEXF_ANISOTROPIC;
    MipFilter     = D3DTEXF_LINEAR;
    MipMapLodBias = -1;
};

sampler Water_EnvMapSampler = sampler_state
{
    Texture       = (Water_EnvTexture);
    AddressU      = Clamp;
    AddressV      = Clamp;
    MinFilter     = (State_MinFilter);
    MagFilter     = (State_MagFilter);
    MipFilter     = (State_MipFilter);
    MipMapLodBias = (State_MipMapLodBias);
};

struct VSWATER20_INPUT
{
    float4 vPos       : POSITION;
    float4 vColor     : COLOR;
    float3 vNormal    : NORMAL;
    float2 vTex       : TEXCOORD0;
};

struct VSWATER20_OUTPUT
{
    float4  vPos      : POSITION;
    float   fFog      : FOG;
    float4  vColor    : COLOR0;

    float4 TexBump        : TEXCOORD0;
    float4 TexEnvMap      : TEXCOORD1;
    float4 TexBase        : TEXCOORD2;
    float4 MiscData       : TEXCOORD3;
    float3 vNormalWS      : TEXCOORD4;
    float3 vPosWS         : TEXCOORD5;
    float3 vEyePosWS      : TEXCOORD6;
    float4 TexScreen      : TEXCOORD7;
};

float4 ComputeTextureCoordinates(float3 WorldPos, float Rotation, float UVScale, float XSpeed, float YSpeed, float Blend, float WindBlend)
{
    float2 bmp,sc;
    float2 tempBump;
    float waveDirection;

    waveDirection = g_fWindDirection + radians(Rotation);
    bmp = fmod((WorldPos.xz - mWorld[3].xz) * (UVScale * 0.0004f),256.0f);
    sc = float2(sin(waveDirection),cos(waveDirection));
    tempBump = bmp + (fSimTime * float2(XSpeed,YSpeed) * 0.003f);
    return float4(tempBump.x * sc.y - tempBump.y * sc.x, tempBump.x * sc.x + tempBump.y * sc.y, Blend * WindBlend, Blend);
}

VSWATER20_OUTPUT Water20VS(const VSWATER20_INPUT v,const bool b30Shader)
{
    VSWATER20_OUTPUT o = (VSWATER20_OUTPUT)0;

    // Transform to clip space
    o.vPos = mul(v.vPos, mFullProj);

    // Transform to world space
        float3 vWorldPos = mul(v.vPos, mWorld);

    // Compute eye vector
        float3 vEye = normalize(vCamEyePoint - vWorldPos);
        
    // Scale normal based bump based on dist
        //float fNormalScale = max(1.0,length(vEye)-fBumpDistScale);

    // Compute bumpmap coordinates
    	float windFactor = 0.9 + (pow(g_fWindSpeed/32.0f,0.5f) * 0.2f);

    	// o.TexBump.xy = ComputeTextureCoordinates( vWorldPos, 141.516, 3.847, -1.283, 1.004, 0.757,windFactor*0.109);
    	// o.TexBump.zw = ComputeTextureCoordinates( vWorldPos, 141.516, 3.847, -1.423, 1.838, 0.751,windFactor*0.609);

    	o.TexBump.xy = ComputeTextureCoordinates( vWorldPos, fBumpRotation, fBumpUVScale, fBumpTimeX1, fBumpTimeY1, 0.757,windFactor*1.000);
    	o.TexBump.zw = ComputeTextureCoordinates( vWorldPos, fBumpRotation, fBumpUVScale, fBumpTimeX2, fBumpTimeY2, 0.751,windFactor*0.609);

        // Send out interpolated normal for lighting
    	o.vNormalWS = normalize(TransformVectorToSpace(v.vNormal, mWorld));

        // Send out eye pos
        o.vPosWS    = vWorldPos;
        o.vEyePosWS = vCamEyePoint;

    	// Output vertex color straight to PS
    	// The alpha component is used to represent the relative
    	// depth of the water, used in computing water translucenty
    	// The color component is simply the color scaled by the 
    	// diffuse sun lighting component
    	// R = N/A
    	// G = Fog
    	// B = Intensity
    	// A = Alpha
    	o.vColor = float4(pow(o.vNormalWS.y,3), ComputeFog(o.vPos.w, FOG_END, FOG_RECIP, FOG_SELECT_VERTEX, FOG_SELECT_TABLE), pow(o.vNormalWS.y,3),v.vColor.a);

    // Environment mapping is determined from the screen-space coordinates
    // of our geometry. This implies that we need to divide them by W, and then
    // offset/scale into a 0-1 range. Because of interpolation issues due to front
    // clipped geometry, the info is pre-processed and passed to the PS where the 1/W
    // is applied. We also apply the normal to the data to emulate reflections caused
    // by the animating geometry

    o.TexEnvMap.xyz = o.vPos.xyz+o.vPos.www;
    o.TexEnvMap.w = 2.0*o.vPos.w;
    o.TexScreen.xyz = float3(1,-1,1)*o.vPos.xyz+o.vPos.www;
    o.TexScreen.w = 2.0*o.vPos.w;

    // Diffuse texture coordinates
    o.TexBase.xy = v.vTex;

    // Compute per-poly fresnel
    o.MiscData =   float4(1.0 - lerp(fFresnelFactorMin, fFresnelFactorMax, pow(smoothstep(0,1,saturate(-normalize(vWorldPos).y)),0.5)),0.844,0,0);

    return o;
}

float4 Water20PS(VSWATER20_OUTPUT Input, const bool b30Shader) : COLOR
{
    float4 Bump;
    float2 BumpEnv;
    float4 Color;
        float2 specularFactor;

    // Renomalize the input vectors and determine the
    // eye vector for lighting and reflection
    float3 vEyeVectWS = Input.vEyePosWS - Input.vPosWS;
    float eyeDist = length(vEyeVectWS);
    float3 vEyeDirWS = (Input.vEyePosWS - Input.vPosWS) / eyeDist;

    // Sample the base color and bumpmaps
    // Since this is essentially a texbem style operation, we
    // manually apply the matrix to the bumpmap values, we then 
    // scale our envmap texcoords by 1/W before factoring in the
    // bumpmap

    Bump = tex2Dbias(Water_BumpSampler, float4(Input.TexBump.xy,0,-2.0f)) + tex2Dbias(Water_BumpSampler, float4(Input.TexBump.zw,0,-3.0f));
    Bump = normalize(Bump-1.0f);

    float3 vSurfaceNorm = Bump.xzy;

    // Compute the diffuse lighting component of the water.
    float lightFactor_sun = saturate(dot(vSunVectorWorld_Sun,vSurfaceNorm));
    float3 vHN = normalize(vEyeDirWS.xyz + vSunVectorWorld_Sun.xyz);
    float3 vHN2 = normalize(vEyeDirWS.xyz + vSunVectorWorld_Moon.xyz);

    specularFactor = fSpecularBlend*(pow(fSpecularBoost*saturate(float2(dot(vSurfaceNorm, vHN),fSunMoonInterpolant * dot(vSurfaceNorm, vHN2))), fSpecularPower));
    
    // Sample the water textures and determine the final color
    BumpEnv = (Input.TexEnvMap.xy/Input.TexEnvMap.w)+ (0.2f*Bump.xy);
    float4 envColor = tex2Dbias(Water_EnvMapSampler, float4(BumpEnv.xy,0,1 * (Input.TexScreen.z/Input.TexScreen.w)));
        
    float4 baseColor = tex2D(Water_BaseSampler, Input.TexBase);
    float fAlpha = baseColor.a;

    float4 terrainColor = baseColor;
    float4 fLand = { baseColor.a, baseColor.a, baseColor.a, baseColor.a };
        
    // Determine the final color
    Color = float4(( (0.8 * lightFactor_sun+0.6) * lerp(baseColor, (0.5+baseColor)*envColor, Input.MiscData.x)).xyz,fAlpha);
    Color = float4(lerp(Color.xyz + (vSunDirectional * specularFactor.x) + (0.3 * (specularFactor.y + saturate(dot(vSunVectorWorld_Moon,vSurfaceNorm)))).xxx, terrainColor.xyz, fLand),fAlpha);
    return float4(FogPS(Color, Input.vColor.g, g_fFogDensity, g_cFog), Color.a);

}

// ------- 3.0 ----------
float4 Water30PS(VSWATER20_OUTPUT Input) : COLOR
{
        return Water20PS(Input,true);
}

// ------- 2.0 ----------


float4 Water20PS(VSWATER20_OUTPUT Input) : COLOR
{
        return Water20PS(Input,false);
}

// ------ VS 2.0 ------
VSWATER20_OUTPUT Water20VS(VSWATER20_INPUT Input)
{
        return Water20VS(Input,false);
}

// ------ VS 3.0 ------
VSWATER20_OUTPUT Water30VS(VSWATER20_INPUT Input)
{
        return Water20VS(Input,true);
}